#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @created: 20.02.2024
# @author: Aleksey Komissarov
# @contact: ad3002@gmail.com

from intervaltree import IntervalTree
from tqdm import tqdm

def get_bed_coordinates(x, ref2tf, chrm2start):
  try:
    start_header = ref2tf.get_header(x.begin).split()[0]
    end_header = ref2tf.get_header(x.end).split()[0]
    start_pos = x.begin - chrm2start[start_header]
    end_pos = x.end - chrm2start[end_header]
  except:
    print("ERROR2:", x)
    return None, None, None
  if start_header != end_header:
    print("ERROR1:", start_header, start_pos, end_header, end_pos, rep)
    return None, None, None
  if end_pos < start_pos:
      start_pos, end_pos = end_pos, start_pos
  return start_header, start_pos, end_pos


def get_bed_data(all_loci, ref2tf, chrm2start, locus_length_cutoff=10000):

  data = []
  for rep in all_loci:
    for x in all_loci[rep]:
      start_header, start_pos, end_pos = get_bed_coordinates(x, ref2tf, chrm2start)
      if start_header is None:
        continue
      if abs(end_pos - start_pos) < locus_length_cutoff:
        continue
      data.append((start_header, start_pos, end_pos, rep))

  return data


def compute_stats(data, trf_our_format, name="our"):
  ref_IT = {}
  for chrmA, startA, endA, repA in trf_our_format:
    ref_IT.setdefault(chrmA, IntervalTree())
    ref_IT[chrmA].addi(startA, endA, repA)

  dat_IT = {}
  for chrmA, startA, endA, repA in data:
    dat_IT.setdefault(chrmA, IntervalTree())
    dat_IT[chrmA].addi(startA, endA, repA)

  false_positive = 0
  false_negative =0
  true_positive = 0
  true_negative = 0
  true_positive_our = 0

  missed_repeats_fp = []
  missed_repeats_fn = []

  for chrmA, startA, endA, repA in trf_our_format:
    if chrmA in dat_IT and dat_IT[chrmA][startA:endA]:
      true_positive += 1
    else:
      false_negative += 1
      missed_repeats_fn.append((chrmA, startA, endA, repA))


  for chrmA, startA, endA, repA in data:
    if chrmA in ref_IT and not ref_IT[chrmA][startA:endA]:
      false_positive += 1
      missed_repeats_fp.append((chrmA, startA, endA, repA))
    else:
      true_positive_our += 1


  print("FP", false_positive)
  print("FN", false_negative)
  print("TP", true_positive)
  print("TN", true_negative)
  print(f"Accuracy {name} to trf:", true_positive/len(trf_our_format))
  print(f"Accuracy trf to {name}:", true_positive_our/len(data))
  p = true_positive/(true_positive+false_positive)
  r = true_positive/(true_positive+false_negative)
  print("P", p)
  print("R", r)
  print("F1", 2 * p * r / (p+r))

  evaluation = {
    "FP": false_positive,
    "FN": false_negative,
    "TP": true_positive,
    "TN": true_negative,
    "Accuracy": true_positive/len(trf_our_format),
    "P": p,
    "R": r,
    "F1": 2 * p * r / (p+r)
  }

  return evaluation, missed_repeats_fp, missed_repeats_fn


### version 2 all kmers
### get all ref loci for repeats

def compute_loci(predicted_trs, ref2tf, delta=1000, k=23):
  ### TODO: replace interval tree with chains
  all_loci = {}
  N = len(predicted_trs)
  for rep_id, repeat in enumerate(predicted_trs):
    computed_kmers = set()
    if len(repeat) < k:
      repeat = (repeat * k)[:2*k]
    pos_tree = IntervalTree()
    print(f"{rep_id}/{N}", repeat)
    for i in tqdm(range(len(repeat)-k+1)):
      key_kmer = repeat[i:i+k]
      if key_kmer in computed_kmers:
        continue
      computed_kmers.add(key_kmer)
      ref_poses = ref2tf.pos(key_kmer)
      for pos in ref_poses:
        pos_tree.addi(pos-delta, pos+delta)
    print(len(pos_tree))
    pos_tree.merge_overlaps()
    print(len(pos_tree))
    all_loci[repeat] = pos_tree
  return all_loci




def compute_loci_chains(predicted_trs, ref2tf, delta=1000, min_array_length=10000, k=23, min_fish_strength=100):
  ### TODO: replace interval tree with chains
  all_loci = {}
  N = len(predicted_trs)
  for rep_id, repeat in enumerate(predicted_trs):
    computed_kmers = set()
    if len(repeat) < k:
      repeat = (repeat * k)[:2*k]
    all_positions = []
    print(f"{rep_id}/{N}", repeat)
    for i in tqdm(range(len(repeat)-k+1)):
      key_kmer = repeat[i:i+k]
      if key_kmer in computed_kmers:
        continue
      computed_kmers.add(key_kmer)
      all_positions += ref2tf.pos(key_kmer)
    all_positions.sort()

    chains = []
    i, j = 0, 1
    fish_strength = 0
    while i < len(all_positions) and j < len(all_positions):
      dist = all_positions[j] - all_positions[i]
      if dist < delta:
        j += 1
        fish_strength += 1
      else:
        if fish_strength > min_fish_strength:
          d = all_positions[j-1] - all_positions[i]
          if d >= min_array_length:
            chains.append((all_positions[i], all_positions[j-1]))
        i = j
        j = i + 1
        fish_strength = 0
    
    if fish_strength > min_fish_strength and i < len(all_positions):
      d = all_positions[j-1] - all_positions[i]
      if d >= min_array_length:
        chains.append((all_positions[i], all_positions[j-1]))
    ## todo: handle last case
    pos_tree = IntervalTree()
    for i, j in chains:
      pos_tree.addi(i, j)
    print(len(pos_tree))
    all_loci[repeat] = pos_tree
  return all_loci


def compute_score(monomers_dataset, trf_our_format, chrm2start, ref2tf, delta=30_000, min_array_length=100,min_fish_strength=100, locus_length_cutoff=10_000, k=23):
    all_loci_chains = compute_loci_chains(monomers_dataset, ref2tf, delta=delta, min_array_length=min_array_length, k=k, min_fish_strength=min_fish_strength)
    data = get_bed_data(all_loci_chains, ref2tf, chrm2start, locus_length_cutoff=locus_length_cutoff)
    evaluation, missed_repeats_fp, missed_repeats_fn = compute_stats(data, trf_our_format)
    return evaluation, missed_repeats_fp, missed_repeats_fn


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @created: 20.02.2024
# @author: Aleksey Komissarov
# @contact: ad3002@gmail.com

from collections import defaultdict
from .helpers import get_revcomp
from tqdm import tqdm
from collections import deque

def naive_tr_finder(sdat, kmer2tf, min_tf_extension=3000, min_fraction_to_continue=30, k=23):
    ### TODO: save forks decisions

    repeats = []
    used_kmers = set()

    alphabet = ["A", "C", "T", "G"]
    rid = 0
    kmer2rid = {}
    kmer2repeat = defaultdict(list)
    frag2type = {}
    for (start_kmer, tf) in sdat:
        step = 0
        if start_kmer in used_kmers:
            continue
        used_kmers.add(start_kmer)
        used_kmers.add(get_revcomp(start_kmer))
        frag2type[start_kmer] = "SS"
        frag2type[get_revcomp(start_kmer)] = "SS"
        kmer2rid[start_kmer] = rid
        seq = start_kmer
        profile = [tf]
    while True:
        kmer = seq[-k:]
        ori_tf = kmer2tf[kmer]
        solutions = []
        prefix = kmer[-k+1:]
        for i, nucleotide in enumerate(alphabet):
            ctf = kmer2tf[prefix+nucleotide]
            if ctf < min_tf_extension:
                continue
            fr = 100.*ctf//(ori_tf+1)
            solutions.append((fr, nucleotide))

        solutions.sort()
    
        if solutions and solutions[-1][0] >= min_fraction_to_continue:
            seq += "".join(solutions[-1][1])
            profile.append(solutions)
            step += 1

            if seq[-k:] in used_kmers and seq[-k:] != start_kmer:
                repeated_kmer = seq[-k:]
                prev_type = frag2type[repeated_kmer]

                if len(prev_type) == 2:
                    prev_type = "F"+prev_type

                repeats.append((rid, "FRAG", len(seq), start_kmer, seq))
                for iii in range(len(seq)-k+1):
                    kmer_ = seq[iii:iii+k]
                    kmer2repeat[kmer_].append((rid, iii, prev_type))
                    kmer2repeat[get_revcomp(kmer_)].append((rid, iii, prev_type))
                    if not kmer_ in frag2type or frag2type[kmer_] == "FPP":
                        frag2type[kmer_] = prev_type
                        frag2type[get_revcomp(kmer_)] = prev_type
                rid += 1
                break
            used_kmers.add(seq[-k:])
            used_kmers.add(get_revcomp(seq[-k:]))
            frag2type[seq[-k:]] = "PP"
            frag2type[get_revcomp(seq[-k:])] = "PP"
            kmer2rid[seq[-k:]] = rid
            ### TODO: case ABAC => AB
            if seq[-k:] == start_kmer:
                repeats.append((rid, "TR", len(seq[:step]), seq[:step], profile))
                for iii in range(len(seq)-k+1):
                    kmer2repeat[seq[iii:iii+k]].append((rid, iii, "TR"))
                    kmer2repeat[get_revcomp(seq[iii:iii+k])].append((rid, iii, "TR"))
                    frag2type[seq[iii:iii+k]] = "TR"
                    frag2type[get_revcomp(seq[iii:iii+k])] = "TR"
                rid += 1
                break

            continue
        repeats.append((rid, "TE", len(seq), seq, solutions))
        for iii in range(len(seq)-k+1):
            kmer2repeat[seq[iii:iii+k]].append((rid, iii, "TE"))
            kmer2repeat[get_revcomp(seq[iii:iii+k])].append((rid, iii, "TE"))
            frag2type[seq[iii:iii+k]] = "TE"
            frag2type[get_revcomp(seq[iii:iii+k])] = "TE"
        rid += 1
        break

    return repeats, kmer2rid, kmer2repeat, frag2type

def tr_greedy_finder(sdat, kmer2tf, max_depth=30_000, coverage=30, min_fraction_to_continue=30, k=23):
    # ### Step 5b. Greedy find the most possible circles in the graph not working

    MIN_TF = coverage * 100

    repeats = []
    rid = 0

    alphabet = ["A", "C", "T", "G"]
    rid = 0
    cache = {}
    print("Expected iterations:", len([x for x in sdat if x[1] > MIN_TF]))
    for (start_kmer, tf) in tqdm(sdat):
        if tf < MIN_TF:
            break
        if start_kmer in cache:
            continue
        second_status = None
        status = None
        next_rid = None
        next_i = None
        length = k
        seq = [start_kmer]
        prefix = start_kmer[1:]

        cache[start_kmer] = (rid, 0, length)

        while True:
            solutions = []
            for i, nucleotide in enumerate(alphabet):
                ctf = kmer2tf[prefix+nucleotide]
                if ctf < MIN_TF:
                    continue
                if ctf:
                    solutions.append((ctf, nucleotide))
            if not solutions:
                status = "zero"
                break

            solutions.sort()
            ctf, nucleotide = solutions[-1]
            kmer = prefix + nucleotide
            seq.append(nucleotide)
            if start_kmer == kmer:
                status = "tr"
                break

            prefix = kmer[1:]
            length += 1

            if kmer in cache:
                rid, strand, i = cache[kmer]
                if not rid in repeats:
                    second_status = "self"
                else:
                    second_status = repeats[rid][0]
                next_rid = rid
                next_i = i
                status = "frag"
                # seq.append(repeats[rid][-1][i:])
                break
            cache[kmer] = (rid, 0, length)
            if length == max_depth:
                status == "long"
                break
        repeats.append((status, second_status, next_rid, next_i, "".join(seq)))
        rid += 1

    return repeats

def tr_greedy_finder_bidirectional(sdat, kmer2tf, max_depth=30_000, coverage=30, min_fraction_to_continue=30, k=23):
    """
    Extends sequences in both left and right directions except for TRs.

    Parameters:
        sdat: List of tuples (kmer, tf) sorted by tf descending.
        kmer2tf: Dictionary mapping k-mer to its tf value.
        max_depth: Maximum length to extend.
        coverage: Coverage threshold.
        min_fraction_to_continue: Minimum fraction to continue extending.
        k: Length of k-mers.

    Returns:
        List of repeats with their status and sequences.
    """

    MIN_TF = coverage * 100

    repeats = []
    rid = 0

    alphabet = ["A", "C", "T", "G"]
    cache = {}
    print("Expected iterations:", len([x for x in sdat if x[1] > MIN_TF]))

    for (start_kmer, tf) in tqdm(sdat):
        if tf < MIN_TF:
            break
        if start_kmer in cache:
            continue

        # Initialize for bidirectional extension
        seq_deque = deque([start_kmer])
        final_status = None
        second_status = None
        next_rid = None
        next_i = None
        total_length = k

        # Mark the start k-mer in cache
        cache[start_kmer] = (rid, 0, total_length)

        # Right extension
        right_prefix = start_kmer[1:]
        right_seq = []
        right_length = 0
        right_status = None

        while right_length < max_depth:
            solutions = []
            for nucleotide in alphabet:
                new_kmer = right_prefix + nucleotide
                ctf = kmer2tf[new_kmer]
                if ctf < MIN_TF:
                    continue
                if ctf:
                    solutions.append((ctf, nucleotide))
            if not solutions:
                right_status = "zero"
                break

            # Select the nucleotide with the highest coverage
            solutions.sort(reverse=True)
            ctf, nucleotide = solutions[0]
            new_kmer = right_prefix + nucleotide
            right_seq.append(nucleotide)
            total_length += 1

            if new_kmer == start_kmer:
                right_status = "tr"
                print(f"TR detected during right extension for k-mer {start_kmer}")
                break

            if new_kmer in cache:
                existing_rid, strand, i = cache[new_kmer]
                if existing_rid not in repeats:
                    second_status = "self"
                else:
                    second_status = repeats[existing_rid][0]
                next_rid = existing_rid
                next_i = i
                right_status = "frag"
                break

            # Update for next iteration
            cache[new_kmer] = (rid, 0, total_length)
            right_prefix = new_kmer[1:]
            right_length += 1

        # If TR detected in right extension, skip left extension
        if right_status == "tr":
            final_status = "tr"
            # To include the start_kmer, reconstruct the full sequence
            # Since it's a TR, the sequence loops back to the start_kmer
            # The full sequence should represent one full repeat unit
            # Assuming right_seq represents the repeat unit after the start_kmer
            # For simplicity, concatenate start_kmer and right_seq
            # Adjust this logic based on the actual TR structure
            full_seq = start_kmer + ''.join(right_seq)
        else:
            # Left extension
            left_suffix = start_kmer[:-1]
            left_seq = []
            left_length = 0
            left_status = None

            while left_length < max_depth:
                solutions = []
                for nucleotide in alphabet:
                    new_kmer = nucleotide + left_suffix
                    ctf = kmer2tf[new_kmer]
                    if ctf < MIN_TF:
                        continue
                    if ctf:
                        solutions.append((ctf, nucleotide))
                if not solutions:
                    left_status = "zero"
                    break

                # Select the nucleotide with the highest coverage
                solutions.sort(reverse=True)
                ctf, nucleotide = solutions[0]
                new_kmer = nucleotide + left_suffix
                left_seq.append(nucleotide)
                total_length += 1

                if new_kmer == start_kmer:
                    left_status = "tr"
                    print(f"TR detected during left extension for k-mer {start_kmer}")
                    break

                if new_kmer in cache:
                    existing_rid, strand, i = cache[new_kmer]
                    if existing_rid not in repeats:
                        second_status = "self"
                    else:
                        second_status = repeats[existing_rid][0]
                    next_rid = existing_rid
                    next_i = i
                    left_status = "frag"
                    break

                # Update for next iteration
                cache[new_kmer] = (rid, 0, total_length)
                left_suffix = new_kmer[:-1]
                left_length += 1

            # Determine overall status
            if left_status == "tr" or right_status == "tr":
                final_status = "tr"
            elif left_status == "frag" or right_status == "frag":
                final_status = "frag"
            elif left_status == "zero" and right_status == "zero":
                final_status = "zero"
            else:
                final_status = "extended"

            # Combine left and right sequences
            # Reverse left_seq to get the correct order
            full_seq = ''.join(reversed(left_seq)) + start_kmer + ''.join(right_seq)

        # Handle cases where full_seq might still be empty
        if not full_seq:
            print(f"Warning: Empty sequence for repeat ID {rid}, start_kmer {start_kmer}")
            full_seq = None

        repeats.append((final_status, second_status, next_rid, next_i, full_seq))
        rid += 1

    return repeats


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @created: 20.02.2024
# @author: Aleksey Komissarov
# @contact: ad3002@gmail.com

from .helpers import get_revcomp

def load_sdat_as_dict(file_name, minimal_tf=100):
    ''' Load kmer tf data from sorted tsv file.
    '''
    data = {}
    with open(file_name) as fh:
        for line in fh:
            kmer, tf = line.strip().split()
            data[kmer] = int(tf)
            if int(tf) < minimal_tf:
                break
    return data

def load_sdat_as_list(file_name, minimal_tf=100):
    ''' Load kmer tf data from sorted tsv file.
    '''
    data = []
    with open(file_name) as fh:
        for line in fh:
            kmer, tf = line.strip().split()
            data.append((kmer, int(tf)))
            if int(tf) < minimal_tf:
                break
    return data

def compute_abundace_anomaly(sdat, ref_sdat, coverage, ref_coverage=1):
    ''' Compute abundance anomaly between two sdat files.
    '''
    all_rep = []
    kmer2abandacy_diff = {}
    for kmer, tf in sdat:
        a = tf//coverage
        b = 0
        if kmer in ref_sdat:
            b = ref_sdat[kmer]//ref_coverage
        v = a/(b+0.000001)
        all_rep.append((v, kmer, a, b))
        kmer2abandacy_diff[kmer] = (v, kmer, a, b)
        kmer2abandacy_diff[get_revcomp(kmer)] = (v, kmer, a, b)
    all_rep.sort(reverse=True)
    return all_rep, kmer2abandacy_diff



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @created: 20.02.2024
# @author: Aleksey Komissarov
# @contact: ad3002@gmail.com

import argparse

from extractr.core_functions.sdat_tools import load_sdat_as_dict
from .core_functions.index_tools import compute_and_get_index
import aindex
from .core_functions.sdat_tools import compute_abundace_anomaly
from .core_functions.tr_finder import naive_tr_finder, tr_greedy_finder
from collections import Counter
from trseeker.tools.sequence_tools import get_revcomp
from .core_functions.evaluation import compute_score
from trseeker.tools.sequence_tools import sc_iter_fasta_brute


def get_ref_aindex(fasta_file, ref_prefix, threads, lu=1):
    
    settings = {
        "index_prefix": "/media/eternus1/data/human/raw_reads/GCF_009914755.1_T2T-CHM13v2.0_genomic.L2.23",
        "aindex_prefix": "/media/eternus1/data/human/raw_reads/GCF_009914755.1_T2T-CHM13v2.0_genomic.L2.23",
        "reads_file": "/media/eternus1/data/human/raw_reads/GCF_009914755.1_T2T-CHM13v2.0_genomic.L2.reads",
        "max_tf": 10000000,
    }

    ref2tf = aindex.load_aindex(settings, skip_reads=False, skip_aindex=False)
    ref2tf.load_header("/media/eternus1/data/human/raw_reads/GCF_009914755.1_T2T-CHM13v2.0_genomic.L2.header")

    chrm2start = {}
    for start, name in ref2tf.headers.items():
        chrm2start[name.split()[0]] = start

    sdat_file = f"{ref_prefix}.sdat"
    sdat_file = "/media/eternus1/data/human/raw_reads/GCF_009914755.1_T2T-CHM13v2.0_genomic.L2.23"

    sdat_ref = load_sdat_as_dict(sdat_file, minimal_tf=lu)

    trf_file = "/media/eternus1/nfs/projects/databases/t2t/trf/GCF_009914755.1_T2T-CHM13v2.0_genomic.10kb.trf"
    ground_truth = {}
    trf_data = []
    trf_our_format = []
    with open(trf_file) as fh:
        for line in tqdm(fh):
            d = line.strip().split("\t")
            trf_data.append(d)
            array = d[14].upper()
            if len(array) < 100000:
                continue
            # for i in range(len(array)-k+1):
            #   kmer = array[i:i+k]
            #   ground_truth.setdefault(kmer, {})
            #   trid = len(trf_data) - 1
            #   ground_truth[kmer].setdefault(trid, 0)
            #   ground_truth[kmer][trid] += 1
            trf_our_format.append((d[18].split()[0], int(d[6]), int(d[7]), d[13].upper()))
    
    return ref2tf, chrm2start, sdat_ref, trf_our_format


def run_it(settings):

    ### step 1. Compute aindex for reads

    fastq1 = settings["fastq1"]
    fastq2 = settings["fastq2"]
    output = settings["output"]
    threads = settings.get("threads", 32)
    coverage = settings.get("coverage", 1)
    lu = settings.get("lu", 100 * coverage)
    prefix = settings.get("index", "raw")
    fasta_file = settings.get("fasta", None)
    ref_prefix = settings.get("rindex", "ref")

    kmer2tf, sdat = compute_and_get_index(fastq1, fastq2, prefix, threads, lu=lu)
    ref2tf, chrm2start, sdat_ref, trf_our_format = get_ref_aindex(fasta_file, ref_prefix, threads, lu=1)

    ### Step 4a. Find kmers underrepresented in the assembly
    ### Step 4b. Find kmers overrepresented in the assembly

    all_rep, kmer2abandacy_diff = compute_abundace_anomaly(sdat, sdat_ref, coverage, ref_coverage=1)

    absent_in_ref = [x for x in all_rep if x[3] == 0]
    absent_in_raw = [x for x in all_rep if x[2] == 0]

    result_unerrepresented_file = f"{output}_underrepresented.csv"
    with open(result_unerrepresented_file, "w") as fh:
        for rep in absent_in_ref:
            fh.write(f"{'\t'.join(map(str,rep))}\n")

    result_overrepresented_file = f"{output}_overrepresented.csv"
    with open(result_overrepresented_file, "w") as fh:
        for rep in absent_in_raw:
            fh.write(f"{'\t'.join(map(str,rep))}\n")

    ### step 2. Find tandem repeats using circular path in de bruijn graph

    naive_tr_repeats, _, _, _ = naive_tr_finder(sdat, kmer2tf, min_tf_extension=3000, min_fraction_to_continue=30, k=23)
    repeats = tr_greedy_finder(sdat, kmer2tf, max_depth=30_000, coverage=30, min_fraction_to_continue=30, k=23)
    print(Counter([x[0] for x in repeats]))

    all_predicted_trs_v2 = []
    for i, (status, second_status, next_rid, next_i, seq) in enumerate(repeats):
        if status == "tr":
            seq = seq[:-k]
            print(status, second_status, next_rid, next_i, len(seq), seq)
            all_predicted_trs_v2.append(seq)
    print("Len of all_predicted_trs_v2", len(all_predicted_trs_v2))

    ### 6. Get dataset for evaluation

    MIN_PRED_SIZE = 30000

    trii = 0
    all_predicted_all = []
    for ii, x in enumerate(repeats):
        trii += 1
        all_predicted_all.append(x[3])
        all_predicted_all.append(get_revcomp(x[3]))

    trii = 0
    all_predicted_trs = []
    for ii, x in enumerate(repeats):
        if x[1] == "TR":
            trii += 1
            monomer = x[3]
            if len(monomer) < 5:
                continue
            kmer = (x[3] * k)[:k]
            size = kmer2tf[kmer]//30 * len(x[3])
            if size < MIN_PRED_SIZE:
                continue
            print(trii, f"Size: {size}", x)
            # all_predicted_trs.append((size, x[3]))
            all_predicted_trs.append(x[3])
            # all_predicted_trs.append(get_revcomp(x[3]))
    print(len(all_predicted_trs))

    evaluation1, missed_repeats_fp1, missed_repeats_fn1 = compute_score(all_predicted_trs_v2, trf_our_format, chrm2start, ref2tf, delta=30_000, min_array_length=100, min_fish_strength=100, locus_length_cutoff=10_000, k=23)
    evaluation2, missed_repeats_fp2, missed_repeats_fn2 = compute_score(all_predicted_trs, trf_our_format, chrm2start, ref2tf, delta=30_000, min_array_length=100, min_fish_strength=100, locus_length_cutoff=10_000, k=23)

    srf_file = "/mnt/data/human/raw_reads/srf.fa"

    srf_reps = []
    for header, seq in sc_iter_fasta_brute(srf_file):
        if isinstance(seq, list):
            seq = seq[0]
        if len(seq) > 15000:
            continue
        srf_reps.append(seq)
    print(len(srf_reps))

    evaluation_srf, missed_repeats_fp_srf, missed_repeats_fn_srf = compute_score(srf_reps, trf_our_format, chrm2start, ref2tf, delta=30_000, min_array_length=100, min_fish_strength=100, locus_length_cutoff=10_000, k=23)

    ### step 3. Save evaluation results to CSV

    output_file = f"{output}.csv"

    
    evaluation_fields = ["FP", "FN", "TP", "TN", "Accuracy", "P", "R", "F1"]

    with open(output_file, "w") as fh:
        for field in evaluation_fields:
            fh.write(f"{field},")
        fh.write("\n")
        for field in evaluation_fields:
            fh.write(f"{evaluation1[field]},")
        fh.write("\n")
        for field in evaluation_fields:
            fh.write(f"{evaluation2[field]},")
        fh.write("\n")
        for field in evaluation_fields:
            fh.write(f"{evaluation_srf[field]},")
        fh.write("\n")


    found_repeats_fasta = f"{output}.fasta"
    with open(found_repeats_fasta, "w") as fh:
        for i, seq in enumerate(repeats):
            fh.write(f">{i}_{len(seq)}bp\n{seq}\n")
        

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Benchmark on human T2T data.")
    parser.add_argument("--fastq1", help="Input left reads in FASTQ format.")
    parser.add_argument("--fastq2", help="Input right reads in  in FASTQ format.")
    parser.add_argument("-o", "--output", help="Output prefix for tandem repeats search results.")
    parser.add_argument("--rindex", help="Index prefix for reference.")
    parser.add_argument("--index", help="Index prefix for raw reads.")
    parser.add_argument("--fasta", help="Reference data in fasta format.")
    parser.add_argument("-t", "--threads", help="Number of threads to use.")
    parser.add_argument("-c", "--coverage", help="Coverage for aindex.")
    parser.add_argument("--lu", help="LU for aindex")
    parser.add_argument("--hl", help="Heng Li fasta")
    parser.add_argument("--tarean", help="TAREAM fasta")

    args = parser.parse_args()

    settings = {
        "fastq1": args.fastq1,
        "fastq2": args.fastq2,
        "output": args.output,
        "threads": args.threads,
        "coverage": args.coverage,
        "lu": args.lu,
        "rindex": args.rindex,
        "index": args.index,
        "fasta": args.fasta,
        "hl": args.hl,
        "tarean": args.tarean,
    }

    run_it(settings)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @created: 20.02.2024
# @author: Aleksey Komissarov
# @contact: ad3002@gmail.com

import os
from .sdat_tools import load_sdat_as_list
import aindex

def compute_and_get_index(fastq1, fastq2, prefix, threads, lu=2):

    sdat_file = f"{prefix}.23.sdat"
    index_prefix_file = f"{prefix}.23"

    if not os.path.isfile(sdat_file) or  not os.path.isfile(index_prefix_file):
        command = f"python3 compute_aindex.py -i {fastq1},{fastq2} -t fastq -o {prefix} --lu {lu} --sort 1 -P {threads} --onlyindex 1"
        print(command)
        os.system(command)

    sdat = load_sdat_as_list(sdat_file, minimal_tf=lu)

    ### Step 2. Load raw reads aindex

    settings = {
        "index_prefix": f"{prefix}.23",
        "aindex_prefix": f"{prefix}.23",
        "reads_file": f"{prefix}.reads",
        "max_tf": 10000000,
    }

    kmer2tf = aindex.load_aindex(settings, skip_reads=True, skip_aindex=True)
    return kmer2tf, sdat

def print_kmer_right_read_fragments(kmer, kmer2tf, read_length=310, topk=100, split_springs=True):
  pos = kmer2tf.pos(kmer)
  for p in pos[:topk]:
    if split_springs:
      print(kmer2tf.reads[p:p+read_length].split(b"\n")[0].split("~")[0])
    else:
      print(kmer2tf.reads[p:p+read_length].split(b"\n")[0])

def print_kmer_left_read_fragments(kmer, kmer2tf, topk=100, read_length=310, k=23, split_springs=True):
  pos = kmer2tf.pos(kmer)
  for p in pos[:topk]:
    if split_springs:
      print(kmer2tf.reads[p-read_length:p+k].split(b"\n")[-1].split(b"~")[-1])
    else:
      print(kmer2tf.reads[p-read_length:p+k].split(b"\n")[-1])


def get_kmer_right_read_fragments(kmer, kmer2tf, read_length=310, topk=100, split_springs=True):
  pos = kmer2tf.pos(kmer)
  reads = []
  for p in pos[:topk]:
    if split_springs:
      reads.append(kmer2tf.reads[p:p+read_length].split(b"\n")[0].split(b"~")[0])
    else:
      reads.append(kmer2tf.reads[p:p+read_length].split(b"\n")[0])
  return reads

def print_kmer_left_read_fragments(kmer, kmer2tf, topk=100, read_length=310, k=23, split_springs=True):
  pos = kmer2tf.pos(kmer)
  reads = []
  for p in pos[:topk]:
    if split_springs:
      reads.append(kmer2tf.reads[p-read_length:p+k].split(b"\n")[-1].split(b"~")[-1])
    else:
      reads.append(kmer2tf.reads[p-read_length:p+k].split(b"\n")[-1])
  return reads



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @created: 20.02.2024
# @author: Aleksey Komissarov
# @contact: ad3002@gmail.com

import argparse
from .core_functions.index_tools import compute_and_get_index
from .core_functions.tr_finder import tr_greedy_finder_bidirectional

def run_it():

    parser = argparse.ArgumentParser(description="Extract and analyze tandem repeats from raw DNA sequences.")
    parser.add_argument("-1", "--fastq1", help="Input file with DNA sequences in FASTQ format.")
    parser.add_argument("-2", "--fastq2", help="Input file with DNA sequences in FASTQ format.")
    parser.add_argument("-f", "--fasta", help="Input genome fasta file", required=False, default=None)
    parser.add_argument("-o", "output", help="Output file with tandem repeats in CSV format.")
    parser.add_argument("-t", "--threads", help="Number of threads to use.", default=32, type=int, required=False)
    parser.add_argument("-c", "--coverage", help="Coverage to use for aindex.", default=1, type=int, required=False)
    parser.add_argument("-k", "--k", help="K-mer size to use for aindex.", default=23, type=int, required=False)
    args = parser.parse_args()
    
    settings = {
        "fastq1": args.fastq1,
        "fastq2": args.fastq2,
        "fasta": args.fasta,
        "output": args.output,
        "threads": 32,
        "coverage": 1,
        "k": 23,
    }

    fastq1 = settings.get("fastq1", None)
    fastq2 = settings.get("fastq2", None)
    threads = settings.get("threads", 32)
    coverage = settings.get("coverage", 1)
    lu = settings.get("lu", 100 * coverage)
    prefix = settings.get("output", "test")
    k = settings.get("k", 23)

    ### step 1. Compute aindex for reads
    kmer2tf, sdat = compute_and_get_index(fastq1, fastq2, prefix, threads, lu=lu)

    ### step 2. Find tandem repeats using circular path in de bruijn graph

    repeats = tr_greedy_finder_bidirectional(sdat, kmer2tf, max_depth=30_000, coverage=30, min_fraction_to_continue=30, k=23)

    all_predicted_trs = []
    for i, (status, second_status, next_rid, next_i, seq) in enumerate(repeats):
        if status == "tr":
            seq = seq[:-k]
            print(status, second_status, next_rid, next_i, len(seq), seq)
            all_predicted_trs.append(seq)
        elif status == "frag":
            pass
        elif status == "zero":
            pass
        elif status == "long":
            pass
        else:
            print(status, second_status, next_rid, next_i, len(seq), seq)
            raise Exception("Unknown status")
        
    print(len(all_predicted_trs))

    ### step 3. Save results to CSV

    output_file = f"{prefix}.csv"

    with open(output_file, "w") as fh:
        for i, seq in enumerate(all_predicted_trs):
            fh.write(f">{i}_{len(seq)}bp\n{seq}\n")

    ### step 4. Analyze repeat borders

    ### step 5. Enrich repeats variants
    
if __name__ == "__main__":
    run_it()


import re
from setuptools import find_packages, setup
import os

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

# Попытка прочитать requirements.txt, если он существует
if os.path.exists('requirements.txt'):
    with open('requirements.txt') as f:
        requirements = f.read().splitlines()
else:
    # Если файл не найден, используем минимальный набор зависимостей
    requirements = [
        "tqdm",
        "intervaltree",
        "aindex2",
    ]

version = "0.2.8"

setup(
    name="extracTR",
    version=version,
    packages=find_packages('src'),
    package_dir={'': 'src'},
    package_data={
        "": ["README.md"],
        "extractr": ["requirements.txt"],  # Включаем requirements.txt в пакет
    },
    python_requires=">=3.6",
    include_package_data=True,
    scripts=[],
    license="BSD",
    url="https://github.com/aglabx/extracTR",
    author="Aleksey Komissarov",
    author_email="ad3002@gmail.com",
    description="Extract and analyze satellite DNA from raw sequences.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    entry_points={
        'console_scripts': [
            'extracTR = extractr.extractr:run_it',
        ],
    },
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @created: 20.02.2024
# @author: Aleksey Komissarov
# @contact: ad3002@gmail.com


def get_revcomp(sequence: str) -> str:
    complement = str.maketrans('ATCGNatcgn~[]', 'TAGCNtagcn~][')
    return sequence.translate(complement)[::-1]